from qc_utils.qcmetric import QCMetric, QCMetricRecord

__title__ = "qc-utils"
__version__ = "20.05.1"
__description__ = "Tool for representing ENCODE QC metrics"
__url__ = "https://github.com/ENCODE-DCC/qc-utils"
__uri__ = __url__
__author__ = "Otto Jolanki"
__email__ = "ojolanki@stanford.edu"
__license__ = "MIT"
__copyright__ = "Copyright (c) 2020 ENCODE DCC"
